"""
fortress/query_visitor.py
Ian Kollipara
2022.05.02

Query Visitor Class Definition
"""

# Imports
from ast import (
    AST,
    Add,
    And,
    Attribute,
    BinOp,
    BitAnd,
    BitOr,
    BitXor,
    BoolOp,
    Call,
    Compare,
    Constant,
    Div,
    Eq,
    FloorDiv,
    Gt,
    GtE,
    Invert,
    In,
    NotIn,
    LShift,
    Lambda,
    Lt,
    LtE,
    MatMult,
    Mod,
    Mult,
    Name,
    NodeVisitor,
    Not,
    NotEq,
    Pow,
    RShift,
    Slice,
    Sub,
    Subscript,
    Tuple,
    UAdd,
    USub,
    UnaryOp,
    keyword,
)
from typing import Any, Callable, Dict


class QueryVisitor(NodeVisitor):

    """Query Visitor

    This class subclasses ast.NodeVisitor. Its purpose is to
    walk the newly created AST and create the Deta Base query
    from it.
    """

    OP_TABLE = {Eq: "", NotEq: "?ne", Lt: "?lt", LtE: "?lte", Gt: "?gt", GtE: "?gte"}
    UOP_TABLE = {UAdd: "+", USub: "-", Invert: "~", Not: "not"}
    BOP_TABLE = {
        Pow: "**",
        Mult: "*",
        MatMult: "@",
        FloorDiv: "//",
        Div: "/",
        Mod: "%",
        Add: "+",
        Sub: "-",
        LShift: "<<",
        RShift: ">>",
        BitAnd: "&",
        BitOr: "|",
        BitXor: "^",
    }

    def __init__(
        self, query_ast: AST, deta_base_name: str, globals: Dict[str, Any]
    ) -> None:
        self.db_name = deta_base_name
        self.globals = globals
        self.ast = query_ast
        self.param = ""

    def generate_query(self) -> str:
        return self.visit(self.ast)

    def visit_Lambda(self, node: Lambda) -> Any:
        """visit the Lambda Node.

        This function visits the lambda node, and
        takes note of the function parameters, which
        is replaced by self.db_name.
        """

        self.param = node.args.args[0].arg
        return f"{{{self.visit(node.body)}}}"

    def visit_BoolOp(self, node: BoolOp) -> Any:
        """Visit the BoolOp Node.

        This function visits the bool operation node,
        and will generate the different query based on
        whether the function is an AND or an OR.
        """

        if isinstance(node.op, And):
            queries = [f"{self.visit(child)}" for child in node.values]

        else:
            queries = [f"{{{self.visit(child)}}}" for child in node.values]

        return ", ".join(queries)

    def visit_Attribute(self, node: Attribute) -> Any:
        """Visit the Attribute Node.

        Attributes recurse to their children, thus
        creating the correct string for object's values.
        """

        obj = self.visit(node.value)

        if not isinstance(obj, str):
            return getattr(obj, node.attr)

        return f"{obj}.{node.attr}"

    def visit_Name(self, node: Name) -> Any:
        """Visit the Name Node.

        This node can be two things. If the
        value is the param then just return
        the id value. Otherwise attempt to resolve
        the value through the globals dict.
        """

        if node.id == self.param:
            return self.db_name

        elif resolved_val := self.globals.get(node.id):
            if isinstance(resolved_val, str):
                return f"'{resolved_val}'"

            return resolved_val

        else:
            return node.id

    def visit_Constant(self, node: Constant) -> Any:
        """Visit the Constant Node."""

        if isinstance(node.value, str):
            return f"'{node.value}'"
        return node.value

    def visit_Compare(self, node: Compare) -> Any:
        """Visit the Compare Node."""

        if isinstance(node.left, Subscript):
            return self.handle_prefix(node)

        elif isinstance(node.ops[0], (In, NotIn)):
            return self.handle_contains(node)

        else:
            operator = self.OP_TABLE[node.ops[0].__class__]  # type: ignore
            lhs = self.visit(node.left)
            rhs = self.visit(node.comparators[0])
            return f"'{lhs}{operator}': {rhs}"

    def handle_prefix(self, node: Compare):
        """Handle the issues that occur with prefix query.

        To do a prefix query, the list access syntax may
        be used.
        """

        value = self.visit(node.left.value)  # type: ignore
        compare = self.visit(node.comparators[0])

        return f"'{value}?pfx': {compare}"

    def handle_contains(self, node: Compare):
        """Handle issues with contains in Deta.

        Given that deta cannot do `x.name in [1,2,3]`, the
        adjusted query is created with an OR query. Unless
        the the IN is against a string, which works fine.
        """

        lhs = self.visit(node.left)
        rhs = self.visit(node.comparators[0])

        if isinstance(rhs, range):
            return f"'{lhs}?r': [{rhs.start},{rhs.stop}]"

        if self.db_name in rhs:
            if isinstance(node.ops[0], In):
                return f"'{rhs}?contains': {lhs}"

            return f"'{rhs}?!contains': {lhs}"

        else:
            if isinstance(node.ops[0], In):
                orred_queries = [f"{{'{lhs}': {val}}}" for val in rhs]
            else:
                orred_queries = [f"'{lhs}?ne': {val}" for val in rhs]

            return ", ".join(orred_queries)

    def visit_Call(self, node: Call) -> Any:
        """Visit the Call Node.

        This node represents function calls. Since
        Deta only supports two specific function
        calls: range and previous, those will be
        turned into their correct queries, the
        others will be executed and the evaluated
        result is used.
        """

        func = self.visit(node.func)

        keyword_pairs = [self.visit(keyword) for keyword in node.keywords]
        keywords = {key: value for key, value in keyword_pairs}
        args = [self.visit(arg) for arg in node.args]

        if isinstance(func, Callable):
            return func(*args, **keywords)  # type: ignore
        elif isinstance(func, str):
            if resolved_func := getattr(self.globals["__builtins__"], func, None):
                return resolved_func(*args, **keywords)
            else:
                raise NotImplementedError("Method calls on table class not supported")

    def visit_Subscript(self, node: Subscript) -> Any:
        """Visit the Subscript Node.

        This node handles cases like dictionary access
        and list slicing.
        """

        value = self.visit(node.value)

        match node.slice:

            case Slice(lower, upper, step):
                return eval(f"{value}[{lower}:{upper}:{step}]")

            case _:
                slice = self.visit(node.slice)
                return eval(f"{value}[{slice}]")

    def visit_Slice(self, node: Slice) -> Any:
        """Visit the Slice node.

        This handles the case of list slicing.
        """

        return node.lower, node.upper, node.step

    def visit_keyword(self, node: keyword) -> Any:
        """Visit the keyword Node.

        This takes keywords, and creates tuples of (key, value)
        from.
        """

        return (node.arg, self.visit(node.value))

    def visit_Tuple(self, node: Tuple) -> Any:
        """Visit the Tuple Node.

        This builds a tuple from the elements,
        and evaluates it into an actual tuple.
        """

        return [self.visit(el) for el in node.elts]

    def visit_UnaryOp(self, node: UnaryOp) -> Any:
        """Visit the UnaryOp Node.

        Given the list of unary operators, and
        the fact there's not correlation to the queries,
        it's vital that those operators are evaluated
        against their operand.
        """

        if not isinstance(node.op, Not):
            return eval(f"{self.UOP_TABLE[node.op.__class__]}{self.visit(node.operand)}")  # type: ignore
        else:
            return eval(f"not {self.visit(node.operand)}")

    def visit_BinOp(self, node: BinOp) -> Any:
        """Visit the BinOp Node.

        Given that list of binary operators,
        and the fact there's no direct correlation to
        queries, these values must be evaluated.
        """

        operator = self.BOP_TABLE[node.op.__class__]  # type: ignore
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        return eval(f"{lhs} {operator} {rhs}")



"""
fortress/fortress.py
Ian Kollipara
2022.05.05

Fortress Main Class
"""

# Imports
from dataclasses import dataclass, field
from functools import singledispatchmethod
from typing_extensions import Self
from uuid import uuid4
from deta import Deta, _Base
from typing import Any, Callable, ClassVar, Dict, List, Optional, Type, TypeVar
from .decompile import decompile
from .query_visitor import QueryVisitor
from re import compile as re_compile


class Elysium:

    """ Elysium

    Elysium is a Deta ORM. To create a model, subclass the Fortress.Model
    and use the dataclass decorator. This allows for model querying using
    the lambda syntax.
    """

    _models: List[Type["Elysium.Model"]] = []
    _bases: Dict[Type["Elysium.Model"], _Base] = {}

    def __init__(self, deta: Optional[Deta] = None) -> None:
        self.deta = deta or Deta()


    def create_mappings(self):
        """Generate bases for registered models."""

        self._bases = {model: self.deta.Base(model.__name__) for model in self._models}


    @classmethod
    def insert(cls, item: "Elysium.Model" | List["Fortress.Model"]):
        """Insert one or more items into the deta base."""

        base = cls._bases[cls] # type: ignore

        if isinstance(item, list):
            item_dicts = [i.__dict__ for i in item]
            base.put_many(item_dicts) # type: ignore

        else:
            item_dict = item.__dict__.copy()
            key = item_dict.pop("key")
            base.put(item_dict, key)


    @dataclass
    class Model:

        """ Elysium Model

        This is the base class for all Elysium Models. It defines
        how a model is implemented and stored in Elysium.

        To use simply subclass it as Elysium.Model. Ideally the
        model would also be a dataclass. When subclassing, you
        may provide an alternative table name, as well as a callable
        to generate keys from. These keys should be no argument functions
        that result in a string.
        """

        __camel_to_snake = re_compile(r"(?<!^)(?=[A-Z])")
        __table_name__: ClassVar[str]
        __key_func: ClassVar[Callable[[Self], str]]


        def __init_subclass__(cls, table_name: Optional[str] = None, key_generator: Optional[Callable[[], str]] = None) -> None:
            cls.__table_name__ = table_name or cls.__camel_to_snake.sub("_", cls.__name__).lower()
            cls.__key_func = (lambda x: key_generator()) if key_generator else (lambda x: str(uuid4()))

            Elysium._models.append(cls)

        def __post_init__(self):
            self.key = self.__key_func() # type: ignore


        @classmethod
        def fetch(cls, query: Optional[Callable[[Self], bool]] = None, *, limit: int = 1000, offset: int = 0) -> List[Self]:
            """ Fetch a collection of models from the deta base.

            Fetching can be filtered by writing a lambda function to serve as the query. This
            function should take only the class as a parameter, and return a boolean. The amount
            of results can be limited by the limit parameter. You may also offset your results
            through an offset parameter.
            """

            query_str = None if not query else eval(QueryVisitor(decompile(query), cls.__table_name__, query.__globals__).generate_query()) # type: ignore

            items: List[Self] = []

            base = Elysium._bases[cls]
            response = base.fetch(query_str, limit=1000)
            items += [cls(**item) for item in response.items]

            if limit > 1000:
                leftover = limit - 1000

                while leftover <= 0 or response.last is not None:
                    response = base.fetch(query_str, limit=1000)
                    items += [cls(**item) for item in response.items]
                    leftover -= 1000

            return items

        @classmethod
        def get(cls, key: str) -> Self | None:
            """Get an item by its key."""

            if base := Elysium._bases.get(cls):
                if item := base.get(key):
                    return cls(**item)

        @classmethod
        def delete(cls, key: str):
            """Delete the item at the given key."""

            Elysium._bases[cls].delete(key)

        @classmethod
        def update(cls, updated_item: Self):
            """Update the item with the new values."""

            base = Elysium._bases[cls]
            update_dict = updated_item.__dict__.copy()
            key = update_dict.pop("key")
            base.put(update_dict, key)



'''
Fortress is a Typed ODM for Deta's Base NoSQL Database.

The wrapper exists to aid in testing and code reliability.
It uses dataclass objects to map to Bases. This allows
one to query the bases with these models. It also includes
a verbose testing suite to aid in development.
'''

# Exports
from .elysium import Elysium



"""
decompile.py
Ian Kollipara
2022.04.25

Python Bytecode to AST
"""

# Imports
import ast
import dis
import types
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar
from collections.abc import Container
from .query_visitor import QueryVisitor

T = TypeVar("T")


class _Decompiler:
    """_Decompiler is the internal decompiler class.

    _Decompiler is used to take the lambda function queries
    and decompile their bytecode to an AST. This is done to
    preserve strong type checking and a more pythonic way of
    writing queries. This decompiler is designed exclusively
    for use in Fortress, meaning it only decompiles statements
    that fortress cares about: lambdas that result in booleans.

    This class should never be instantiated by the end user.
    """

    compare_operators = {
        "==": ast.Eq(),
        ">": ast.Gt(),
        ">=": ast.GtE(),
        "<=": ast.LtE(),
        "<": ast.Lt(),
        "!=": ast.NotEq(),
    }

    def __init__(self, lambda_func: Callable[[Any], bool]) -> None:
        self.function = lambda_func
        self.bytecode = dis.Bytecode(lambda_func)
        self.stack: List[ast.AST] = []
        self.jumps: List[dis.Instruction] = []
        self.ast: ast.AST = ast.Constant(value=None)
        self.lambda_param = ""

    def __handle_jumps(self, instruction: dis.Instruction):
        """Handle Jump Cases in the Bytecode.

        Given the instance of a jump command, evaluate whether
        the jump occurred due to an And or Or statement.
        """

        jump_argvals = (jump.argval for jump in self.jumps)

        # In the bytecode if a statement is a jump statement, the
        # line to jump to is stored as the argval in the instruction.
        # Thus, to check if a statement is a jump to statement and
        # adjust the stack accordingly, we check if the instruction
        # offset is equal to any jump argvals.
        if instruction.offset in jump_argvals:
            jump: dis.Instruction = next(
                filter(lambda jump: instruction.offset == jump.argval, self.jumps)
            )
            operator = (
                ast.Or()
                if jump.opname in ("JUMP_IF_TRUE_OR_POP", "POP_JUMP_IF_FALSE")
                else ast.And()
            )

            rhs = self.stack.pop()
            lhs = self.stack.pop()
            self.stack.append(ast.BoolOp(op=operator, values=[lhs, rhs]))
            self.jumps.remove(jump)

    def decompile(self):
        """Decompile the provided function to its AST.

        This is the main function of the _Decompiler class. This
        reads each instruction and generates the AST node from it.
        All nodes are stored on the stack, and the resulting AST
        is stored under self.ast.
        """

        for instruction in self.bytecode:

            self.__handle_jumps(instruction)

            try:
                method = getattr(self, instruction.opname.lower())

                method(instruction)
            except AttributeError as e:
                raise NotImplementedError("Unsupported Bytecode") from e

        body = self.stack.pop()
        lambda_node = ast.Lambda(
            args=ast.arguments(
                posonlyargs=[],
                args=[ast.arg(arg=self.lambda_param)],
                kwonlyargs=[],
                kwdefault=[],
                default=[],
            ),
            body=body,
            type_ignore=[],
        )

        self.ast = lambda_node
        return lambda_node

    def return_value(self, instruction: dis.Instruction):
        """Decompile the RETURN_VALUE instruction.

        This instruction doesn't matter as there's need to know
        what the function returns, that's inferred from the AST.
        """

        pass

    def load_attr(self, instruction: dis.Instruction):
        """Decompile the LOAD_ATTR instruction."""

        value = self.stack.pop()
        attr = ast.Attribute(value=value, attr=instruction.argval, ctx=ast.Load())
        self.stack.append(attr)

    def load_fast(self, instruction: dis.Instruction):
        """Decompile the LOAD_FAST instruction.

        Because LOAD_FAST is only used on the
        lambda's parameter, the lambda parameter
        can be inferred from this instruction's argval.
        """

        self.lambda_param = instruction.argval
        self.stack.append(ast.Name(id=instruction.argval, ctx=ast.Load()))

    def load_global(self, instruction: dis.Instruction):
        """Decompile the LOAD_GLOBAL instruction."""

        self.stack.append(ast.Name(id=instruction.argval, ctx=ast.Load()))

    def load_const(self, instruction: dis.Instruction):
        """Decompile the LOAD_CONST instruction."""

        if (
            not isinstance(instruction.argval, str)
            and isinstance(instruction.argval, Container)
            and isinstance(instruction.argval, Iterable)
        ):
            self.stack.append(
                ast.Tuple(elts=[ast.Constant(value=arg) for arg in instruction.argval])
            )
        else:
            self.stack.append(ast.Constant(value=instruction.argval))

    def call_function(self, instruction: dis.Instruction):
        """Decompile the CALL_FUNCTION instruction."""

        args = [self.stack.pop() for _ in range(instruction.argval)]
        func = self.stack.pop()

        call_op = ast.Call(func=func, args=args, keywords=[])
        self.stack.append(call_op)

    def call_function_kw(self, instruction: dis.Instruction):
        """Decompile the CALL_FUNCTION_KW instruction."""

        arg_names = [node.value for node in self.stack.pop().elts]  # type: ignore
        args = [self.stack.pop() for _ in range(instruction.argval)]
        func = self.stack.pop()

        keywords = [
            ast.keyword(arg=arg, value=value) for arg, value in zip(arg_names, args)
        ]

        try:
            unnamed_args = args[len(keywords) :]
        except IndexError:
            unnamed_args = []

        call_op = ast.Call(func=func, args=unnamed_args, keywords=keywords)

        self.stack.append(call_op)

    def build_slice(self, instruction: dis.Instruction):
        """Decompile the BUILD_SLICE instruction."""

        if instruction.argval == 3:
            step = self.stack.pop()
            upper = self.stack.pop()
            lower = self.stack.pop()

            slice_op = ast.Slice(lower=lower, upper=upper, step=step)
        else:
            upper = self.stack.pop()
            lower = self.stack.pop()

            slice_op = ast.Slice(lower=lower, upper=upper)

        self.stack.append(slice_op)

    def load_method(self, instruction: dis.Instruction):
        """Decompile the LOAD_METHOD instruction."""

        value = self.stack.pop()
        attr_op = ast.Attribute(value=value, attr=instruction.argval, ctx=ast.Load())
        self.stack.append(attr_op)

    def call_method(self, instruction: dis.Instruction):
        """Decompile the CALL_METHOD instruction."""

        args = [self.stack.pop() for _ in range(instruction.argval)]
        func = self.stack.pop()

        call_op = ast.Call(func=func, args=args, keywords=[])
        self.stack.append(call_op)

    def unary_positive(self, instruction: dis.Instruction):
        """Decompile the UNARY_POSITIVE instruction."""

        operand = self.stack.pop()
        unary_positive = ast.UnaryOp(op=ast.UAdd(), operand=operand, type_ignores=[])
        self.stack.append(unary_positive)

    def unary_negative(self, instruction: dis.Instruction):
        """Decompile the UNARY_POSITIVE instruction."""

        operand = self.stack.pop()
        unary_negative = ast.UnaryOp(op=ast.USub(), operand=operand, type_ignores=[])
        self.stack.append(unary_negative)

    def unary_not(self, instruction: dis.Instruction):
        """Decompile the UNARY_NOT instruction."""

        operand = self.stack.pop()
        unary_not = ast.UnaryOp(op=ast.Not(), operand=operand, type_ignores=[])
        self.stack.append(unary_not)

    def unary_invert(self, instruction: dis.Instruction):
        """Decompile the UNARY_NOT instruction."""

        operand = self.stack.pop()
        unary_invert = ast.UnaryOp(op=ast.Invert(), operand=operand, type_ignores=[])
        self.stack.append(unary_invert)

    def binary_power(self, instruction: dis.Instruction):
        """Decompile the BINARY_POWER instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        pow_op = ast.BinOp(left=lhs, op=ast.Pow(), right=rhs)
        self.stack.append(pow_op)

    def binary_multiply(self, instruction: dis.Instruction):
        """Decompile the BINARY_MULTIPLY instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        mult_op = ast.BinOp(left=lhs, op=ast.Mult(), right=rhs)
        self.stack.append(mult_op)

    def binary_matrix_multiply(self, instruction: dis.Instruction):
        """Decompile the BINARY_MATRIX_MULTIPLY instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        mat_mult_op = ast.BinOp(left=lhs, op=ast.MatMult(), right=rhs)
        self.stack.append(mat_mult_op)

    def binary_floor_divide(self, instruction: dis.Instruction):
        """Decompile the BINARY_FLOOR_DIVIDE instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        floor_div_op = ast.BinOp(left=lhs, op=ast.FloorDiv(), right=rhs)
        self.stack.append(floor_div_op)

    def binary_true_divide(self, instruction: dis.Instruction):
        """Decompile the BINARY_TRUE_DIVIDE instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        true_div_op = ast.BinOp(left=lhs, op=ast.Div(), right=rhs)
        self.stack.append(true_div_op)

    def binary_modulo(self, instruction: dis.Instruction):
        """Decompile the BINARY_MODULO instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        mod_op = ast.BinOp(left=lhs, op=ast.Mod(), right=rhs)
        self.stack.append(mod_op)

    def binary_add(self, instruction: dis.Instruction):
        """Decompile the BINARY_ADD instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        add_op = ast.BinOp(left=lhs, op=ast.Add(), right=rhs)
        self.stack.append(add_op)

    def binary_subtract(self, instruction: dis.Instruction):
        """Decompile the BINARY_SUBTRACT instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        sub_op = ast.BinOp(left=lhs, op=ast.Sub(), right=rhs)
        self.stack.append(sub_op)

    def binary_subscr(self, instruction: dis.Instruction):
        """Decompile the BINARY_SUBSCR instruction."""

        slice = self.stack.pop()
        value = self.stack.pop()

        subscr_op = ast.Subscript(value=value, slice=slice)
        self.stack.append(subscr_op)

    def binary_lshift(self, instruction: dis.Instruction):
        """Decompile the BINARY_LSHIFT instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        lshift_op = ast.BinOp(left=lhs, op=ast.LShift(), right=rhs)
        self.stack.append(lshift_op)

    def binary_rshift(self, instruction: dis.Instruction):
        """Decompile the BINARY_RSHIFT instruction."""

        lhs = self.stack.pop()
        rhs = self.stack.pop()

        rshift_op = ast.BinOp(left=lhs, op=ast.RShift(), right=rhs)
        self.stack.append(rshift_op)

    def binary_and(self, instruction: dis.Instruction):
        """Decompile the BINARY_ADD instruction."""

        lhs = self.stack.pop()
        rhs = self.stack.pop()

        bit_and_op = ast.BinOp(left=lhs, op=ast.BitAnd(), right=rhs)
        self.stack.append(bit_and_op)

    def binary_xor(self, instruction: dis.Instruction):
        """Decompile the BINARY_XOR instruction."""

        lhs = self.stack.pop()
        rhs = self.stack.pop()

        bit_xor_op = ast.BinOp(left=lhs, op=ast.BitXor(), right=rhs)
        self.stack.append(bit_xor_op)

    def binary_or(self, instruction: dis.Instruction):
        """Decompile the BINARY_OR instruction."""

        lhs = self.stack.pop()
        rhs = self.stack.pop()

        bit_or_op = ast.BinOp(left=lhs, op=ast.BitOr(), right=rhs)
        self.stack.append(bit_or_op)

    def build_tuple(self, instruction: dis.Instruction):
        """Decompile the BUILD_TUPLE instruction."""

        tuple_op = ast.Tuple(elts=[self.stack.pop() for _ in range(instruction.argval)])
        self.stack.append(tuple_op)

    def build_map(self, instruction: dis.Instruction):
        """Decompile the BUILD_MAP instruction."""

        # Pairs are stored as Value, Key in the tree.
        pairs = [
            (self.stack.pop(), self.stack.pop()) for _ in range(instruction.argval)
        ]
        keys = [p[1] for p in pairs]
        values = [p[0] for p in pairs]

        map_op = ast.Dict(keys=keys, values=values)
        self.stack.append(map_op)

    def build_const_key_map(self, instruction: dis.Instruction):
        """Decompile the BUILD_CONST_KEY_MAP instruction."""

        keys = self.stack.pop().elts  # type: ignore
        values = [self.stack.pop() for _ in range(instruction.argval)]

        map_op = ast.Dict(keys=keys, values=values)
        self.stack.append(map_op)

    def is_op(self, instruction: dis.Instruction):
        """Decompile the IS_OP instruction."""

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        operator = ast.IsNot() if instruction.argval else ast.Is()

        compare_op = ast.Compare(left=lhs, ops=[operator], comparators=[rhs])

        self.stack.append(compare_op)

    def compare_op(self, instruction: dis.Instruction):
        """Decompile the COMPARE_OP instruction.

        This handles all comparision cases, which is why
        the class attribute compare_operators exists.
        """

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        compare_op = ast.Compare(
            left=lhs,
            ops=[self.compare_operators[instruction.argval]],
            comparators=[rhs],
        )

        self.stack.append(compare_op)

    def contains_op(self, instruction: dis.Instruction):
        """Decompile the CONTAINS_OP instruction.

        This handles uses of `in` and `not in`.
        """

        rhs = self.stack.pop()
        lhs = self.stack.pop()

        # If the instruction is "not in" then the
        # instruction's argval is set to 1, that is True.
        operator = ast.NotIn() if instruction.argval else ast.In()

        # Contains is really just specialized comparision
        compare_op = ast.Compare(left=lhs, ops=[operator], comparators=[rhs])
        self.stack.append(compare_op)

    def jump_if_false_or_pop(self, instruction: dis.Instruction):
        """Decompile the JUMP_IF_FALSE_OR_POP instruction.

        Since this handles a jump, its actual understanding
        is handled within self.__handle_jump.
        """

        self.jumps.append(instruction)

    def jump_if_true_or_pop(self, instruction: dis.Instruction):
        """Decompile the JUMP_IF_TRUE_OR_POP instruction.

        Since this handles a jump, its actual understanding
        is handled within self.__handle_jump.
        """

        self.jumps.append(instruction)

    def pop_jump_if_true(self, instruction: dis.Instruction):
        """Decompile the POP_JUMP_IF_TRUE instruction.

        Since this handles a jump, its actual understanding
        is handled within self.__handle_jump.
        """

        self.jumps.append(instruction)

    def pop_jump_if_false(self, instruction: dis.Instruction):
        """Decompile the POP_JUMP_IF_FALSE instruction.

        Since this handles a jump, its actual understanding
        is handled within self.__handle_jump.
        """

        self.jumps.append(instruction)


def decompile(lambda_func: types.LambdaType) -> ast.AST:
    """Decompile the given lambda function into its AST>

    This creates a _Decompiler object and invokes the decompile
    method. This is simply a wrapper.
    """

    return _Decompiler(lambda_func).decompile()
